/*
 * BNG Blaster (BBL) - Configuration
 *
 * Hannes Gredler, July 2020
 * Christian Giese, October 2020
 *
 * Copyright (C) 2020-2021, RtBrick, Inc.
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include "bbl.h"
#include "bbl_config.h"
#include "bbl_stream.h"
#include <jansson.h>
#include <sys/stat.h>

const char g_default_user[] = "user{session-global}@rtbrick.com";
const char g_default_pass[] = "test";
const char g_default_ari[] = "DEU.RTBRICK.{session-global}";
const char g_default_aci[] = "0.0.0.0/0.0.0.0 eth 0:{session-global}";

static void
add_secondary_ipv4(bbl_ctx_s *ctx, uint32_t ipv4) {
    bbl_secondary_ip_s  *secondary_ip;

    if(ipv4 != ctx->config.network_ip) {
        /* Add secondary IP address to be served by ARP */
        secondary_ip = ctx->config.secondary_ip_addresses;
        if(secondary_ip) {
            while(secondary_ip) {
                if(secondary_ip->ip == ipv4) {
                    /* Address is already known ... */
                    break;
                }
                if(secondary_ip->next) {
                    /* Check next address ... */
                    secondary_ip = secondary_ip->next;
                } else {
                    /* Append secondary address ... */
                    secondary_ip->next = malloc(sizeof(bbl_secondary_ip_s));
                    memset(secondary_ip->next, 0x0, sizeof(bbl_secondary_ip_s));
                    secondary_ip = secondary_ip->next;
                    secondary_ip->ip = ipv4;
                    break;
                }
            }
        } else {
            /* Add first secondary address */
            ctx->config.secondary_ip_addresses = malloc(sizeof(bbl_secondary_ip_s));
            memset(ctx->config.secondary_ip_addresses, 0x0, sizeof(bbl_secondary_ip_s));
            ctx->config.secondary_ip_addresses->ip = ipv4;
        }
    }
}

static void
add_secondary_ipv6(bbl_ctx_s *ctx, ipv6addr_t ipv6) {
    bbl_secondary_ip6_s  *secondary_ip6;

    if(memcmp(ipv6, ctx->config.network_ip6.address, IPV6_ADDR_LEN) != 0) {
        /* Add secondary IP address to be served by ICMPv6 */
        secondary_ip6 = ctx->config.secondary_ip6_addresses;
        if(secondary_ip6) {
            while(secondary_ip6) {
                if(memcmp(secondary_ip6->ip, ctx->config.network_ip6.address, IPV6_ADDR_LEN) == 0) {
                    /* Address is already known ... */
                    break;
                }
                if(secondary_ip6->next) {
                    /* Check next address ... */
                    secondary_ip6 = secondary_ip6->next;
                } else {
                    /* Append secondary address ... */
                    secondary_ip6->next = malloc(sizeof(bbl_secondary_ip6_s));
                    memset(secondary_ip6->next, 0x0, sizeof(bbl_secondary_ip6_s));
                    secondary_ip6 = secondary_ip6->next;
                    memcpy(secondary_ip6->ip, ipv6, IPV6_ADDR_LEN);
                    break;
                }
            }
        } else {
            /* Add first secondary address */
            ctx->config.secondary_ip6_addresses = malloc(sizeof(bbl_secondary_ip6_s));
            memset(ctx->config.secondary_ip6_addresses, 0x0, sizeof(bbl_secondary_ip6_s));
            memcpy(ctx->config.secondary_ip6_addresses->ip, ipv6, IPV6_ADDR_LEN);
        }
    }
}

static bool
json_parse_access_line_profile (json_t *config, bbl_access_line_profile_s *profile) {
    json_t *value = NULL;

    value = json_object_get(config, "access-line-profile-id");
    if (value) {
        profile->access_line_profile_id = json_number_value(value);
    } else {
        fprintf(stderr, "Config error: Missing value for access-line-profiles->access-line-profile-id\n");
        return false;
    }

    value = json_object_get(config, "act-up");
    if (value) {
        profile->act_up = json_number_value(value);
    }
    value = json_object_get(config, "act-down");
    if (value) {
        profile->act_down = json_number_value(value);
    }
    value = json_object_get(config, "min-up");
    if (value) {
        profile->min_up = json_number_value(value);
    }
    value = json_object_get(config, "min-down");
    if (value) {
        profile->min_down = json_number_value(value);
    }
    value = json_object_get(config, "att-up");
    if (value) {
        profile->att_up = json_number_value(value);
    }
    value = json_object_get(config, "att-down");
    if (value) {
        profile->att_down = json_number_value(value);
    }
    value = json_object_get(config, "min-up-low");
    if (value) {
        profile->min_up_low = json_number_value(value);
    }
    value = json_object_get(config, "min-down-low");
    if (value) {
        profile->min_down_low = json_number_value(value);
    }
    value = json_object_get(config, "max-interl-delay-up");
    if (value) {
        profile->max_interl_delay_up = json_number_value(value);
    }
    value = json_object_get(config, "act-interl-delay-up");
    if (value) {
        profile->act_interl_delay_up = json_number_value(value);
    }
    value = json_object_get(config, "max-interl-delay-down");
    if (value) {
        profile->max_interl_delay_down = json_number_value(value);
    }
    value = json_object_get(config, "act-interl-delay-down");
    if (value) {
        profile->act_interl_delay_down = json_number_value(value);
    }
    value = json_object_get(config, "data-link-encaps");
    if (value) {
        profile->data_link_encaps = json_number_value(value);
    }
    value = json_object_get(config, "dsl-type");
    if (value) {
        profile->dsl_type = json_number_value(value);
    }
    value = json_object_get(config, "pon-type");
    if (value) {
        profile->pon_type = json_number_value(value);
    }
    value = json_object_get(config, "etr-up");
    if (value) {
        profile->etr_up = json_number_value(value);
    }
    value = json_object_get(config, "etr-down");
    if (value) {
        profile->etr_down = json_number_value(value);
    }
    value = json_object_get(config, "attetr-up");
    if (value) {
        profile->attetr_up = json_number_value(value);
    }
    value = json_object_get(config, "attetr-down");
    if (value) {
        profile->attetr_down = json_number_value(value);
    }
    value = json_object_get(config, "gdr-up");
    if (value) {
        profile->gdr_up = json_number_value(value);
    }
    value = json_object_get(config, "gdr-down");
    if (value) {
        profile->gdr_down = json_number_value(value);
    }
    value = json_object_get(config, "attgdr-up");
    if (value) {
        profile->attgdr_up = json_number_value(value);
    }
    value = json_object_get(config, "attgdr-down");
    if (value) {
        profile->attgdr_down = json_number_value(value);
    }
    value = json_object_get(config, "ont-onu-avg-down");
    if (value) {
        profile->ont_onu_avg_down = json_number_value(value);
    }
    value = json_object_get(config, "ont-onu-peak-down");
    if (value) {
        profile->ont_onu_peak_down = json_number_value(value);
    }
    value = json_object_get(config, "ont-onu-max-up");
    if (value) {
        profile->ont_onu_max_up = json_number_value(value);
    }
    value = json_object_get(config, "ont-onu-ass-up");
    if (value) {
        profile->ont_onu_ass_up = json_number_value(value);
    }
    value = json_object_get(config, "pon-max-up");
    if (value) {
        profile->pon_max_up = json_number_value(value);
    }
    value = json_object_get(config, "pon-max-down");
    if (value) {
        profile->pon_max_down = json_number_value(value);
    }

    return true;
}

static bool
json_parse_access_interface (bbl_ctx_s *ctx, json_t *access_interface, bbl_access_config_s *access_config) {
    json_t *value = NULL;
    const char *s = NULL;
    uint32_t ipv4;

    value = json_object_get(access_interface, "i1-start");
    if (value) {
        access_config->i1 = json_number_value(value);
    } else {
        access_config->i1 = 1;
    }
    value = json_object_get(access_interface, "i1-step");
    if (value) {
        access_config->i1_step = json_number_value(value);
    } else {
        access_config->i1_step = 1;
    }
    value = json_object_get(access_interface, "i2-start");
    if (value) {
        access_config->i2 = json_number_value(value);
    } else {
        access_config->i2 = 1;
    }
    value = json_object_get(access_interface, "i2-step");
    if (value) {
        access_config->i2_step = json_number_value(value);
    } else {
        access_config->i2_step = 1;
    }

    if (json_unpack(access_interface, "{s:s}", "type", &s) == 0) {
        if (strcmp(s, "pppoe") == 0) {
            access_config->access_type = ACCESS_TYPE_PPPOE;
        } else if (strcmp(s, "ipoe") == 0) {
            access_config->access_type = ACCESS_TYPE_IPOE;
        } else {
            fprintf(stderr, "JSON config error: Invalid value for access->type\n");
            return false;
        }
    }
    if (json_unpack(access_interface, "{s:s}", "vlan-mode", &s) == 0) {
        if (strcmp(s, "1:1") == 0) {
            access_config->vlan_mode = VLAN_MODE_11;
        } else if (strcmp(s, "N:1") == 0) {
            access_config->vlan_mode = VLAN_MODE_N1;
        } else {
            fprintf(stderr, "JSON config error: Invalid value for access->vlan-mode\n");
            return false;
        }
    }
    if (json_unpack(access_interface, "{s:s}", "interface", &s) == 0) {
        access_config->interface = strdup(s);
    } else {
        fprintf(stderr, "JSON config error: Missing value for access->interface\n");
        return false;
    }

    value = json_object_get(access_interface, "outer-vlan");
    if (json_is_number(value)) {
        access_config->access_outer_vlan_min = json_number_value(value);
        access_config->access_outer_vlan_min &= 4095;
        access_config->access_outer_vlan_max = access_config->access_outer_vlan_min;
    } else {
        value = json_object_get(access_interface, "outer-vlan-min");
        if (json_is_number(value)) {
            access_config->access_outer_vlan_min = json_number_value(value);
            access_config->access_outer_vlan_min &= 4095;
        }
        value = json_object_get(access_interface, "outer-vlan-max");
        if (value) {
            access_config->access_outer_vlan_max = json_number_value(value);
            access_config->access_outer_vlan_max &= 4095;
        }
    }
    value = json_object_get(access_interface, "inner-vlan");
    if (json_is_number(value)) {
        access_config->access_inner_vlan_min = json_number_value(value);
        access_config->access_inner_vlan_min &= 4095;
        access_config->access_inner_vlan_max = access_config->access_inner_vlan_min;
    } else {
        value = json_object_get(access_interface, "inner-vlan-min");
        if (value) {
            access_config->access_inner_vlan_min = json_number_value(value);
            access_config->access_inner_vlan_min &= 4095;
        }
        value = json_object_get(access_interface, "inner-vlan-max");
        if (value) {
            access_config->access_inner_vlan_max = json_number_value(value);
            access_config->access_inner_vlan_max &= 4095;
        }
    }
    if(access_config->access_outer_vlan_min > access_config->access_outer_vlan_max ||
       access_config->access_inner_vlan_min > access_config->access_inner_vlan_max) {
        fprintf(stderr, "JSON config error: Invalid VLAN range (min > max)\n");
        return false;
    }
    value = json_object_get(access_interface, "third-vlan");
    if (value) {
        access_config->access_third_vlan = json_number_value(value);
        access_config->access_third_vlan &= 4095;
    }

    if (json_unpack(access_interface, "{s:s}", "address", &s) == 0) {
        if(!inet_pton(AF_INET, s, &ipv4)) {
            fprintf(stderr, "JSON config error: Invalid value for access->address\n");
            return false;
        }
        access_config->static_ip = ipv4;
    }
    if (json_unpack(access_interface, "{s:s}", "address-iter", &s) == 0) {
        if(!inet_pton(AF_INET, s, &ipv4)) {
            fprintf(stderr, "JSON config error: Invalid value for access->address-iter\n");
            return false;
        }
        access_config->static_ip_iter = ipv4;
    }
    if (json_unpack(access_interface, "{s:s}", "gateway", &s) == 0) {
        if(!inet_pton(AF_INET, s, &ipv4)) {
            fprintf(stderr, "JSON config error: Invalid value for access->gateway\n");
            return false;
        }
        access_config->static_gateway = ipv4;
    }
    if (json_unpack(access_interface, "{s:s}", "gateway-iter", &s) == 0) {
        if(!inet_pton(AF_INET, s, &ipv4)) {
            fprintf(stderr, "JSON config error: Invalid value for access->gateway-iter\n");
            return false;
        }
        access_config->static_gateway_iter = ipv4;
    }

    /* Optionally overload some settings per range */
    if (json_unpack(access_interface, "{s:s}", "username", &s) == 0) {
        access_config->username = strdup(s);
    } else {
        access_config->username = strdup(ctx->config.username);
    }

    if (json_unpack(access_interface, "{s:s}", "password", &s) == 0) {
        access_config->password = strdup(s);
    } else {
        access_config->password = strdup(ctx->config.password);
    }

    if (json_unpack(access_interface, "{s:s}", "authentication-protocol", &s) == 0) {
        if (strcmp(s, "PAP") == 0) {
            access_config->authentication_protocol = PROTOCOL_PAP;
        } else if (strcmp(s, "CHAP") == 0) {
            access_config->authentication_protocol = PROTOCOL_CHAP;
        } else {
            fprintf(stderr, "Config error: Invalid value for access->authentication-protocol\n");
            return false;
        }
    } else {
        access_config->authentication_protocol = ctx->config.authentication_protocol;
    }

    /* Access Line */
    if (json_unpack(access_interface, "{s:s}", "agent-circuit-id", &s) == 0) {
        access_config->agent_circuit_id = strdup(s);
    } else {
        if(ctx->config.agent_circuit_id) {
            access_config->agent_circuit_id = strdup(ctx->config.agent_circuit_id);
        }
    }

    if (json_unpack(access_interface, "{s:s}", "agent-remote-id", &s) == 0) {
        access_config->agent_remote_id = strdup(s);
    } else {
        if(ctx->config.agent_circuit_id) {
            access_config->agent_remote_id = strdup(ctx->config.agent_remote_id);
        }
    }

    value = json_object_get(access_interface, "rate-up");
    if (value) {
        access_config->rate_up = json_number_value(value);
    } else {
        access_config->rate_up = ctx->config.rate_up;
    }

    value = json_object_get(access_interface, "rate-down");
    if (value) {
        access_config->rate_down = json_number_value(value);
    } else {
        access_config->rate_down = ctx->config.rate_down;
    }

    value = json_object_get(access_interface, "dsl-type");
    if (value) {
        access_config->dsl_type = json_number_value(value);
    } else {
        access_config->dsl_type = ctx->config.dsl_type;
    }

    value = json_object_get(access_interface, "access-line-profile-id");
    if (value) {
        access_config->access_line_profile_id = json_number_value(value);
    }

    value = json_object_get(access_interface, "ipcp");
    if (json_is_boolean(value)) {
        access_config->ipcp_enable = json_boolean_value(value);
    } else {
        access_config->ipcp_enable = ctx->config.ipcp_enable;
    }
    value = json_object_get(access_interface, "ip6cp");
    if (json_is_boolean(value)) {
        access_config->ip6cp_enable = json_boolean_value(value);
    } else {
        access_config->ip6cp_enable = ctx->config.ip6cp_enable;
    }
    value = json_object_get(access_interface, "ipv4");
    if (json_is_boolean(value)) {
        access_config->ipv4_enable = json_boolean_value(value);
    } else {
        access_config->ipv4_enable = ctx->config.ipv4_enable;
    }
    value = json_object_get(access_interface, "ipv6");
    if (json_is_boolean(value)) {
        access_config->ipv6_enable = json_boolean_value(value);
    } else {
        access_config->ipv6_enable = ctx->config.ipv6_enable;
    }
    value = json_object_get(access_interface, "dhcp");
    if (json_is_boolean(value)) {
        access_config->dhcp_enable = json_boolean_value(value);
    } else {
        access_config->dhcp_enable = ctx->config.dhcp_enable;
    }
    value = json_object_get(access_interface, "dhcpv6");
    if (json_is_boolean(value)) {
        access_config->dhcpv6_enable = json_boolean_value(value);
    } else {
        access_config->dhcpv6_enable = ctx->config.dhcpv6_enable;
    }
    value = json_object_get(access_interface, "igmp-autostart");
    if (json_is_boolean(value)) {
        access_config->igmp_autostart = json_boolean_value(value);
    } else {
        access_config->igmp_autostart = ctx->config.igmp_autostart;
    }
    value = json_object_get(access_interface, "igmp-version");
    if (json_is_number(value)) {
        access_config->igmp_version = json_number_value(value);
        if(access_config->igmp_version < 1 || access_config->igmp_version > 3) {
            fprintf(stderr, "JSON config error: Invalid value for access->igmp-version\n");
            return false;
        }
    } else {
        access_config->igmp_version = ctx->config.igmp_version;
    }
    value = json_object_get(access_interface, "session-traffic-autostart");
    if (json_is_boolean(value)) {
        access_config->session_traffic_autostart = json_boolean_value(value);
    } else {
        access_config->session_traffic_autostart = ctx->config.session_traffic_autostart;
    }

    value = json_object_get(access_interface, "stream-group-id");
    if (value) {
        access_config->stream_group_id = json_number_value(value);
    }

    value = json_object_get(access_interface, "cfm-cc");
    if (json_is_boolean(value)) {
        access_config->cfm_cc = json_boolean_value(value);
    }
    value = json_object_get(access_interface, "cfm-level");
    if (value) {
        access_config->cfm_level = json_number_value(value);
        if(access_config->cfm_level > 7) {
            fprintf(stderr, "JSON config error: Invalid value for access->cfm-level\n");
            return false;
        }
    }
    value = json_object_get(access_interface, "cfm-ma-id");
    if (value) {
        access_config->cfm_ma_id = json_number_value(value);
    }
    if (json_unpack(access_interface, "{s:s}", "cfm-ma-name", &s) == 0) {
        access_config->cfm_ma_name = strdup(s);
    } else if (access_config->cfm_cc) {
        fprintf(stderr, "JSON config error: Missing access->cfm-ma-name\n");
        return false;
    }

    return true;
}

static bool
json_parse_stream (bbl_ctx_s *ctx, json_t *stream, bbl_stream_config *stream_config) {
    json_t *value = NULL;
    const char *s = NULL;
    double bps;

    if (json_unpack(stream, "{s:s}", "type", &s) == 0) {
        if (strcmp(s, "ipv4") == 0) {
            stream_config->type = STREAM_IPV4;
        } else if (strcmp(s, "ipv6") == 0) {
            stream_config->type = STREAM_IPV6;
        } else if (strcmp(s, "ipv6pd") == 0) {
            stream_config->type = STREAM_IPV6PD;
        } else {
            fprintf(stderr, "JSON config error: Invalid value for stream->type\n");
            return false;
        }
    } else {
        fprintf(stderr, "JSON config error: Missing value for stream->type\n");
        return false;
    }

    if (json_unpack(stream, "{s:s}", "direction", &s) == 0) {
        if (strcmp(s, "upstream") == 0) {
            stream_config->direction = STREAM_DIRECTION_UP;
        } else if (strcmp(s, "downstream") == 0) {
            stream_config->direction = STREAM_DIRECTION_DOWN;
        } else if (strcmp(s, "both") == 0) {
            stream_config->direction = STREAM_DIRECTION_BOTH;
        } else {
            fprintf(stderr, "JSON config error: Invalid value for stream->direction\n");
            return false;
        }
    } else {
        stream_config->direction = STREAM_DIRECTION_BOTH;
        return false;
    }

    if (json_unpack(stream, "{s:s}", "name", &s) == 0) {
        stream_config->name = strdup(s);
    } else {
        fprintf(stderr, "JSON config error: Missing value for stream->name\n");
        return false;
    }

    value = json_object_get(stream, "stream-group-id");
    if (value) {
        stream_config->stream_group_id = json_number_value(value);
    }

    value = json_object_get(stream, "length");
    if (value) {
        stream_config->length = json_number_value(value);
        if(stream_config->length < 76 || stream_config->length > 1500) {
            fprintf(stderr, "JSON config error: Invalid value for stream->length\n");
            return false;
        }
    } else {
        stream_config->length = 128;
    }

    value = json_object_get(stream, "priority");
    if (value) {
        stream_config->priority = json_number_value(value);
    }

    value = json_object_get(stream, "vlan-priority");
    if (value) {
        stream_config->vlan_priority = json_number_value(value);
    }

    value = json_object_get(stream, "pps");
    if (value) {
        stream_config->pps = json_number_value(value);
        if(stream_config->pps == 0) {
            fprintf(stderr, "JSON config error: Invalid value for stream->pps\n");
            return false;
        }
    } else {
        /* pps config has priority over bps */
        value = json_object_get(stream, "bps");
        if (value) {
            bps = json_number_value(value);
            if(!bps) {
                fprintf(stderr, "JSON config error: Invalid value for stream->bps\n");
                return false;
            }
            stream_config->pps = bps / (stream_config->length * 8);
        }
    }
    if(!stream_config->pps) stream_config->pps = 1;

    if (json_unpack(stream, "{s:s}", "network-ipv4-address", &s) == 0) {
        if(!inet_pton(AF_INET, s, &stream_config->ipv4_network_address)) {
            fprintf(stderr, "JSON config error: Invalid value for stream->network-ipv4-address\n");
            return false;
        }
        add_secondary_ipv4(ctx, stream_config->ipv4_network_address);
    }

    if (json_unpack(stream, "{s:s}", "network-ipv6-address", &s) == 0) {
        if(!inet_pton(AF_INET6, s, &stream_config->ipv6_network_address)) {
            fprintf(stderr, "JSON config error: Invalid value for stream->network-ipv6-address\n");
            return false;
        }
        add_secondary_ipv6(ctx, stream_config->ipv6_network_address);
    }

    if (json_unpack(stream, "{s:s}", "destination-ipv4-address", &s) == 0) {
        if(!inet_pton(AF_INET, s, &stream_config->ipv4_destination_address)) {
            fprintf(stderr, "JSON config error: Invalid value for stream->destination-ipv4-address\n");
            return false;
        }
    }

    if (json_unpack(stream, "{s:s}", "destination-ipv6-address", &s) == 0) {
        if(!inet_pton(AF_INET6, s, &stream_config->ipv6_destination_address)) {
            fprintf(stderr, "JSON config error: Invalid value for stream->destination-ipv6-address\n");
            return false;
        }
    }

    value = json_object_get(stream, "threaded");
    if (json_is_boolean(value)) {
        stream_config->threaded = json_boolean_value(value);
    }

    /* Validate configuration */
    if(stream_config->stream_group_id == 0) {
        /* RAW stream */
        if(stream_config->type == STREAM_IPV4) {
            if(!stream_config->ipv4_destination_address) {
                fprintf(stderr, "JSON config error: Missing destination-ipv4-address for RAW stream %s\n", stream_config->name);
                return false;
            }
        }
        if(stream_config->type == STREAM_IPV6) {
            if(!*(uint64_t*)stream_config->ipv6_destination_address) {
                fprintf(stderr, "JSON config error: Missing destination-ipv6-address for RAW stream %s\n", stream_config->name);
                return false;
            }
        }
        if(stream_config->type == STREAM_IPV6PD) {
            fprintf(stderr, "JSON config error: Invalid type for RAW stream %s\n", stream_config->name);
            return false;
        }
        if(stream_config->direction != STREAM_DIRECTION_DOWN) {
            fprintf(stderr, "JSON config error: Invalid direction for RAW stream %s\n", stream_config->name);
            return false;
        }
    }
    return true;
}

static bool
json_parse_config (json_t *root, bbl_ctx_s *ctx) {

    json_t *section, *sub, *value = NULL;
    const char *s;
    uint32_t ipv4;
    int i, size;
    bbl_access_config_s         *access_config          = NULL;
    bbl_access_line_profile_s   *access_line_profile    = NULL;
    bbl_stream_config           *stream_config          = NULL;
    bbl_l2tp_server_t           *l2tp_server            = NULL;

    if(json_typeof(root) != JSON_OBJECT) {
        fprintf(stderr, "JSON config error: Configuration root element must object\n");
        return false;
    }

    /* Sessions Configuration */
    section = json_object_get(root, "sessions");
    if (json_is_object(section)) {
        value = json_object_get(section, "count");
        if (json_is_number(value)) {
            ctx->config.sessions = json_number_value(value);
        }
        value = json_object_get(section, "max-outstanding");
        if (json_is_number(value)) {
            ctx->config.sessions_max_outstanding = json_number_value(value);
        }
        value = json_object_get(section, "start-rate");
        if (json_is_number(value)) {
            ctx->config.sessions_start_rate = json_number_value(value);
        }
        value = json_object_get(section, "stop-rate");
        if (json_is_number(value)) {
            ctx->config.sessions_stop_rate = json_number_value(value);
        }
        value = json_object_get(section, "iterate-vlan-outer");
        if (json_is_boolean(value)) {
            ctx->config.iterate_outer_vlan = json_boolean_value(value);
        }
    }

    /* IPoE Configuration */
    section = json_object_get(root, "ipoe");
    if (json_is_object(section)) {
        value = json_object_get(section, "ipv4");
        if (json_is_boolean(value)) {
            ctx->config.ipv4_enable = json_boolean_value(value);
        }
        value = json_object_get(section, "arp-timeout");
        if (json_is_number(value)) {
            ctx->config.arp_timeout = json_number_value(value);
        }
        value = json_object_get(section, "arp-interval");
        if (json_is_number(value)) {
            ctx->config.arp_interval = json_number_value(value);
        }
        value = json_object_get(section, "ipv6");
        if (json_is_boolean(value)) {
            ctx->config.ipv6_enable = json_boolean_value(value);
        }
    }

    /* PPPoE Configuration */
    section = json_object_get(root, "pppoe");
    if (json_is_object(section)) {
        /* Deprecated ...
         * PPPoE sessions, max-outstanding, start
         * and stop rate was moved to section session
         * as all those values apply to PPPoE and IPoE
         * but for compatibility they are still supported
         * here as well for some time.
         */
        value = json_object_get(section, "sessions");
        if (json_is_number(value)) {
            ctx->config.sessions = json_number_value(value);
        }
        value = json_object_get(section, "max-outstanding");
        if (json_is_number(value)) {
            ctx->config.sessions_max_outstanding = json_number_value(value);
        }
        value = json_object_get(section, "start-rate");
        if (json_is_number(value)) {
            ctx->config.sessions_start_rate = json_number_value(value);
        }
        value = json_object_get(section, "stop-rate");
        if (json_is_number(value)) {
            ctx->config.sessions_stop_rate = json_number_value(value);
        }
        /* ... Deprecated */
        value = json_object_get(section, "session-time");
        if (json_is_number(value)) {
            ctx->config.pppoe_session_time = json_number_value(value);
        }
        value = json_object_get(section, "reconnect");
        if (json_is_boolean(value)) {
            ctx->config.pppoe_reconnect = json_boolean_value(value);
        }
        value = json_object_get(section, "discovery-timeout");
        if (json_is_number(value)) {
            ctx->config.pppoe_discovery_timeout = json_number_value(value);
        }
        value = json_object_get(section, "discovery-retry");
        if (json_is_number(value)) {
            ctx->config.pppoe_discovery_retry = json_number_value(value);
        }
        if (json_unpack(section, "{s:s}", "service-name", &s) == 0) {
            ctx->config.pppoe_service_name = strdup(s);
        }
        value = json_object_get(section, "host-uniq");
        if (json_is_boolean(value)) {
            ctx->config.pppoe_host_uniq = json_boolean_value(value);
        }
        value = json_object_get(section, "vlan-priority");
        if (json_is_number(value)) {
            ctx->config.pppoe_vlan_priority = json_number_value(value);
            if(ctx->config.pppoe_vlan_priority > 7) {
                fprintf(stderr, "JSON config error: Invalid value for pppoe->vlan-priority\n");
                return false;
            }
        }
    }

    /* PPP Configuration */
    section = json_object_get(root, "ppp");
    if (json_is_object(section)) {
        value = json_object_get(section, "mru");
        if (json_is_number(value)) {
            ctx->config.ppp_mru = json_number_value(value);
        }
        sub = json_object_get(section, "authentication");
        if (json_is_object(sub)) {
            if (json_unpack(sub, "{s:s}", "username", &s) == 0) {
                ctx->config.username = strdup(s);
            }
            if (json_unpack(sub, "{s:s}", "password", &s) == 0) {
                ctx->config.password = strdup(s);
            }
            value = json_object_get(sub, "timeout");
            if (json_is_number(value)) {
                ctx->config.authentication_timeout = json_number_value(value);
            }
            value = json_object_get(sub, "retry");
            if (json_is_number(value)) {
                ctx->config.authentication_retry = json_number_value(value);
            }
            if (json_unpack(sub, "{s:s}", "protocol", &s) == 0) {
                if (strcmp(s, "PAP") == 0) {
                    ctx->config.authentication_protocol = PROTOCOL_PAP;
                } else if (strcmp(s, "CHAP") == 0) {
                    ctx->config.authentication_protocol = PROTOCOL_CHAP;
                } else {
                    fprintf(stderr, "JSON config error: Invalid value for ppp->authentication->protocol\n");
                    return false;
                }
            }
        }
        sub = json_object_get(section, "lcp");
        if (json_is_object(sub)) {
            value = json_object_get(sub, "conf-request-timeout");
            if (json_is_number(value)) {
                ctx->config.lcp_conf_request_timeout = json_number_value(value);
            }
            value = json_object_get(sub, "conf-request-retry");
            if (json_is_number(value)) {
                ctx->config.lcp_conf_request_retry = json_number_value(value);
            }
            value = json_object_get(sub, "keepalive-interval");
            if (json_is_number(value)) {
                ctx->config.lcp_keepalive_interval = json_number_value(value);
            }
            value = json_object_get(sub, "keepalive-retry");
            if (json_is_number(value)) {
                ctx->config.lcp_keepalive_retry = json_number_value(value);
            }
            value = json_object_get(sub, "start-delay");
            if (json_is_number(value)) {
                ctx->config.lcp_start_delay = json_number_value(value);
                if(ctx->config.lcp_start_delay >= 1000) {
                    fprintf(stderr, "JSON config error: ppp->lcp->start-delay must be < 1000\n");
                    return false;
                }
            }
        }
        sub = json_object_get(section, "ipcp");
        if (json_is_object(sub)) {
            value = json_object_get(sub, "enable");
            if (json_is_boolean(value)) {
                ctx->config.ipcp_enable = json_boolean_value(value);
            }
            value = json_object_get(sub, "request-ip");
            if (json_is_boolean(value)) {
                ctx->config.ipcp_request_ip = json_boolean_value(value);
            }
            value = json_object_get(sub, "request-dns1");
            if (json_is_boolean(value)) {
                ctx->config.ipcp_request_dns1 = json_boolean_value(value);
            }
            value = json_object_get(sub, "request-dns2");
            if (json_is_boolean(value)) {
                ctx->config.ipcp_request_dns2 = json_boolean_value(value);
            }
            value = json_object_get(sub, "conf-request-timeout");
            if (json_is_number(value)) {
                ctx->config.ipcp_conf_request_timeout = json_number_value(value);
            }
            value = json_object_get(sub, "conf-request-retry");
            if (json_is_number(value)) {
                ctx->config.ipcp_conf_request_retry = json_number_value(value);
            }
        }
        sub = json_object_get(section, "ip6cp");
        if (json_is_object(sub)) {
            value = json_object_get(sub, "enable");
            if (json_is_boolean(value)) {
                ctx->config.ip6cp_enable = json_boolean_value(value);
            }
            value = json_object_get(sub, "conf-request-timeout");
            if (json_is_number(value)) {
                ctx->config.ip6cp_conf_request_timeout = json_number_value(value);
            }
            value = json_object_get(sub, "conf-request-retry");
            if (json_is_number(value)) {
                ctx->config.ip6cp_conf_request_retry = json_number_value(value);
            }
        }
    }

    /* DHCP Configuration */
    section = json_object_get(root, "dhcp");
    if (json_is_object(section)) {
        value = json_object_get(section, "enable");
        if (json_is_boolean(value)) {
            ctx->config.dhcp_enable = json_boolean_value(value);
        }
        value = json_object_get(section, "broadcast");
        if (json_is_boolean(value)) {
            ctx->config.dhcp_broadcast = json_boolean_value(value);
        }
        value = json_object_get(section, "timeout");
        if (json_is_number(value)) {
            ctx->config.dhcp_timeout = json_number_value(value);
        }
        value = json_object_get(section, "retry");
        if (json_is_number(value)) {
            ctx->config.dhcp_retry = json_number_value(value);
        }
        value = json_object_get(section, "release-interval");
        if (json_is_number(value)) {
            ctx->config.dhcp_release_interval = json_number_value(value);
        }
        value = json_object_get(section, "release-retry");
        if (json_is_number(value)) {
            ctx->config.dhcp_release_retry = json_number_value(value);
        }
        value = json_object_get(section, "tos");
        if (json_is_number(value)) {
            ctx->config.dhcp_tos = json_number_value(value);
        }
        value = json_object_get(section, "vlan-priority");
        if (json_is_number(value)) {
            ctx->config.dhcp_vlan_priority = json_number_value(value);
            if(ctx->config.dhcp_vlan_priority > 7) {
                fprintf(stderr, "JSON config error: Invalid value for dhcp->vlan-priority\n");
                return false;
            }
        }
    }

    /* DHCPv6 Configuration */
    section = json_object_get(root, "dhcpv6");
    if (json_is_object(section)) {
        value = json_object_get(section, "enable");
        if (json_is_boolean(value)) {
            ctx->config.dhcpv6_enable = json_boolean_value(value);
        }
        value = json_object_get(section, "rapid-commit");
        if (json_is_boolean(value)) {
            ctx->config.dhcpv6_rapid_commit = json_boolean_value(value);
        }
        value = json_object_get(section, "timeout");
        if (json_is_number(value)) {
            ctx->config.dhcpv6_timeout = json_number_value(value);
        }
        value = json_object_get(section, "retry");
        if (json_is_number(value)) {
            ctx->config.dhcpv6_retry = json_number_value(value);
        }
    }

    /* IGMP Configuration */
    section = json_object_get(root, "igmp");
    if (json_is_object(section)) {
        value = json_object_get(section, "version");
        if (json_is_number(value)) {
            ctx->config.igmp_version = json_number_value(value);
            if(ctx->config.igmp_version < 1 || ctx->config.igmp_version > 3) {
                fprintf(stderr, "JSON config error: Invalid value for igmp->version\n");
                return false;
            }
        }
        value = json_object_get(section, "combined-leave-join");
        if (json_is_boolean(value)) {
            ctx->config.igmp_combined_leave_join = json_boolean_value(value);
        }
        value = json_object_get(section, "autostart");
        if (json_is_boolean(value)) {
            ctx->config.igmp_autostart = json_boolean_value(value);
        }
        value = json_object_get(section, "start-delay");
        if (json_is_number(value) && json_number_value(value)) {
            /* Min 1 second */
            ctx->config.igmp_start_delay = json_number_value(value);
        }
        if (json_unpack(section, "{s:s}", "group", &s) == 0) {
            if(!inet_pton(AF_INET, s, &ipv4)) {
                fprintf(stderr, "JSON config error: Invalid value for igmp->group\n");
                return false;
            }
            ctx->config.igmp_group = ipv4;
        }
        if (json_unpack(section, "{s:s}", "group-iter", &s) == 0) {
            if(!inet_pton(AF_INET, s, &ipv4)) {
                fprintf(stderr, "JSON config error: Invalid value for igmp->group-iter\n");
                return false;
            }
            ctx->config.igmp_group_iter = ipv4;
        }
        if (json_unpack(section, "{s:s}", "source", &s) == 0) {
            if(!inet_pton(AF_INET, s, &ipv4)) {
                fprintf(stderr, "JSON config error: Invalid value for igmp->source\n");
                return false;
            }
            ctx->config.igmp_source = ipv4;
        }
        value = json_object_get(section, "group-count");
        if (json_is_number(value)) {
            ctx->config.igmp_group_count = json_number_value(value);
        }
        value = json_object_get(section, "zapping-interval");
        if (json_is_number(value)) {
            ctx->config.igmp_zap_interval = json_number_value(value);
        }
        value = json_object_get(section, "zapping-view-duration");
        if (json_is_number(value)) {
            ctx->config.igmp_zap_view_duration = json_number_value(value);
        }
        value = json_object_get(section, "zapping-count");
        if (json_is_number(value)) {
            ctx->config.igmp_zap_count = json_number_value(value);
        }
        value = json_object_get(section, "zapping-wait");
        if (json_is_boolean(value)) {
            ctx->config.igmp_zap_wait = json_boolean_value(value);
        }
        value = json_object_get(section, "send-multicast-traffic");
        if (json_is_boolean(value)) {
            ctx->config.send_multicast_traffic = json_boolean_value(value);
        }
        value = json_object_get(section, "multicast-traffic-length");
        if (json_is_number(value)) {
            ctx->config.multicast_traffic_len = json_number_value(value);
            if(ctx->config.multicast_traffic_len > 1500) {
                fprintf(stderr, "JSON config error: Invalid value for igmp->multicast-traffic-length (max 1500)\n");
            }
        }
        value = json_object_get(section, "multicast-traffic-tos");
        if (json_is_number(value)) {
            ctx->config.multicast_traffic_tos = json_number_value(value);
        }
    }

    /* Access Line Configuration */
    section = json_object_get(root, "access-line");
    if (json_is_object(section)) {
        if (json_unpack(section, "{s:s}", "agent-circuit-id", &s) == 0) {
            ctx->config.agent_circuit_id = strdup(s);
        }
        if (json_unpack(section, "{s:s}", "agent-remote-id", &s) == 0) {
            ctx->config.agent_remote_id = strdup(s);
        }
        value = json_object_get(section, "rate-up");
        if (json_is_number(value)) {
            ctx->config.rate_up = json_number_value(value);
        }
        value = json_object_get(section, "rate-down");
        if (json_is_number(value)) {
            ctx->config.rate_down = json_number_value(value);
        }
        value = json_object_get(section, "dsl-type");
        if (json_is_number(value)) {
            ctx->config.dsl_type = json_number_value(value);
        }
    }

    /* Session Traffic Configuration */
    section = json_object_get(root, "session-traffic");
    if (json_is_object(section)) {
        value = json_object_get(section, "autostart");
        if (json_is_boolean(value)) {
            ctx->config.session_traffic_autostart = json_boolean_value(value);
        }
        value = json_object_get(section, "ipv4-pps");
        if (json_is_number(value)) {
            ctx->config.session_traffic_ipv4_pps = json_number_value(value);
        }
        value = json_object_get(section, "ipv6-pps");
        if (json_is_number(value)) {
            ctx->config.session_traffic_ipv6_pps = json_number_value(value);
        }
        value = json_object_get(section, "ipv6pd-pps");
        if (json_is_number(value)) {
            ctx->config.session_traffic_ipv6pd_pps = json_number_value(value);
        }
    }

    /* Interface Configuration */
    section = json_object_get(root, "interfaces");
    if (json_is_object(section)) {
        value = json_object_get(section, "tx-interval");
        if (json_is_number(value)) {
            ctx->config.tx_interval = json_number_value(value) * MSEC;
        }
        value = json_object_get(section, "rx-interval");
        if (json_is_number(value)) {
            ctx->config.rx_interval = json_number_value(value) * MSEC;
        }
        value = json_object_get(section, "qdisc-bypass");
        if (json_is_boolean(value)) {
            ctx->config.qdisc_bypass = json_boolean_value(value);
        }
        value = json_object_get(section, "io-slots");
        if (json_is_number(value)) {
            ctx->config.io_slots = json_number_value(value);
        }
        if (json_unpack(section, "{s:s}", "io-mode", &s) == 0) {
            if (strcmp(s, "packet_mmap_raw") == 0) {
                ctx->config.io_mode = IO_MODE_PACKET_MMAP_RAW;
#if BNGBLASTER_NETMAP
            } else if (strcmp(s, "netmap") == 0) {
                ctx->config.io_mode = IO_MODE_NETMAP;
#endif
            } else if (strcmp(s, "packet_mmap") == 0) {
                ctx->config.io_mode = IO_MODE_PACKET_MMAP;
            } else if (strcmp(s, "raw") == 0) {
                ctx->config.io_mode = IO_MODE_RAW;
            } else {
                fprintf(stderr, "Config error: Invalid value for interfaces->io-mode\n");
                return false;
            }
        } else {
            ctx->config.io_mode = IO_MODE_PACKET_MMAP_RAW;
        }
        value = json_object_get(section, "io-stream-max-ppi");
        if (json_is_number(value)) {
            ctx->config.io_stream_max_ppi = json_number_value(value);
        }
        sub = json_object_get(section, "network");
        if (json_is_object(sub)) {
            if (json_unpack(sub, "{s:s}", "interface", &s) == 0) {
                snprintf(ctx->config.network_if, IFNAMSIZ, "%s", s);
            }
            if (json_unpack(sub, "{s:s}", "address", &s) == 0) {
                if(!inet_pton(AF_INET, s, &ipv4)) {
                    fprintf(stderr, "JSON config error: Invalid value for network->address\n");
                    return false;
                }
                ctx->config.network_ip = ipv4;
            }
            if (json_unpack(sub, "{s:s}", "gateway", &s) == 0) {
                if(!inet_pton(AF_INET, s, &ipv4)) {
                    fprintf(stderr, "JSON config error: Invalid value for network->gateway\n");
                    return false;
                }
                ctx->config.network_gateway = ipv4;
            }
            if (json_unpack(sub, "{s:s}", "address-ipv6", &s) == 0) {
                if(!inet_pton(AF_INET6, s, &ctx->config.network_ip6.address)) {
                    fprintf(stderr, "JSON config error: Invalid value for network->address-ipv6\n");
                    return false;
                }
                ctx->config.network_ip6.len = 64;
            }
            if (json_unpack(sub, "{s:s}", "gateway-ipv6", &s) == 0) {
                if(!inet_pton(AF_INET6, s, &ctx->config.network_gateway6.address)) {
                    fprintf(stderr, "JSON config error: Invalid value for network->gateway-ipv6\n");
                    return false;
                }
                ctx->config.network_gateway6.len = 64;
            }
            value = json_object_get(sub, "vlan");
            if (json_is_number(value)) {
                ctx->config.network_vlan = json_number_value(value);
                ctx->config.network_vlan &= 4095;
            }
        }
        sub = json_object_get(section, "access");
        if (json_is_array(sub)) {
            /* Config is provided as array (multiple access ranges) */
            size = json_array_size(sub);
            for (i = 0; i < size; i++) {
                if(!access_config) {
                    ctx->config.access_config = malloc(sizeof(bbl_access_config_s));
                    access_config = ctx->config.access_config;
                } else {
                    access_config->next = malloc(sizeof(bbl_access_config_s));
                    access_config = access_config->next;
                }
                memset(access_config, 0x0, sizeof(bbl_access_config_s));
                if(!json_parse_access_interface(ctx, json_array_get(sub, i), access_config)) {
                    return false;
                }
            }
        } else if (json_is_object(sub)) {
            /* Config is provided as object (single access range) */
            access_config = malloc(sizeof(bbl_access_config_s));
            memset(access_config, 0x0, sizeof(bbl_access_config_s));
            if(!ctx->config.access_config) {
                ctx->config.access_config = access_config;
            }
            if(!json_parse_access_interface(ctx, sub, access_config)) {
                return false;
            }
        }
    } else {
        fprintf(stderr, "JSON config error: Missing interfaces section\n");
        return false;
    }

    /* L2TP Server Configuration (LNS) */
    section = json_object_get(root, "l2tp-server");
    if (json_is_array(section)) {
        if(!ctx->config.network_gateway || !ctx->config.network_ip) {
            fprintf(stderr, "JSON config error: Failed to add L2TP server because of missing or incomplete network interface config\n");
            return false;
        }
        size = json_array_size(section);
        for (i = 0; i < size; i++) {
            sub = json_array_get(section, i);
            if(!l2tp_server) {
                ctx->config.l2tp_server = malloc(sizeof(bbl_l2tp_server_t));
                l2tp_server = ctx->config.l2tp_server;
            } else {
                l2tp_server->next = malloc(sizeof(bbl_l2tp_server_t));
                l2tp_server = l2tp_server->next;
            }
            memset(l2tp_server, 0x0, sizeof(bbl_l2tp_server_t));
            if (json_unpack(sub, "{s:s}", "name", &s) == 0) {
                l2tp_server->host_name = strdup(s);
            } else {
                fprintf(stderr, "JSON config error: Missing value for l2tp-server->name\n");
                return false;
            }
            if (json_unpack(sub, "{s:s}", "secret", &s) == 0) {
                l2tp_server->secret = strdup(s);
            }
            if (json_unpack(sub, "{s:s}", "address", &s) == 0) {
                if(!inet_pton(AF_INET, s, &ipv4)) {
                    fprintf(stderr, "JSON config error: Invalid value for l2tp-server->address\n");
                    return false;
                }
                l2tp_server->ip = ipv4;
                CIRCLEQ_INIT(&l2tp_server->tunnel_qhead);
                add_secondary_ipv4(ctx, ipv4);
            } else {
                fprintf(stderr, "JSON config error: Missing value for l2tp-server->address\n");
            }
            value = json_object_get(sub, "receive-window-size");
            if (json_is_number(value)) {
                l2tp_server->receive_window = json_number_value(value);
            } else {
                l2tp_server->receive_window = 4;
            }
            value = json_object_get(sub, "max-retry");
            if (json_is_number(value)) {
                l2tp_server->max_retry = json_number_value(value);
            } else {
                l2tp_server->max_retry = 30;
            }
            if (json_unpack(sub, "{s:s}", "congestion-mode", &s) == 0) {
                if (strcmp(s, "default") == 0) {
                    l2tp_server->congestion_mode = BBL_L2TP_CONGESTION_DEFAULT;
                } else if (strcmp(s, "slow") == 0) {
                    l2tp_server->congestion_mode = BBL_L2TP_CONGESTION_SLOW;
                } else if (strcmp(s, "aggressive") == 0) {
                    l2tp_server->congestion_mode = BBL_L2TP_CONGESTION_AGGRESSIVE;
                } else {
                    fprintf(stderr, "Config error: Invalid value for l2tp-server->congestion-mode\n");
                    return false;
                }
            } else {
                l2tp_server->congestion_mode = BBL_L2TP_CONGESTION_DEFAULT;
            }
            value = json_object_get(sub, "data-control-priority");
            if (json_is_boolean(value)) {
                l2tp_server->data_control_priority = json_boolean_value(value);
            }
            value = json_object_get(sub, "data-length");
            if (json_is_boolean(value)) {
                l2tp_server->data_lenght = json_boolean_value(value);
            }
            value = json_object_get(sub, "data-offset");
            if (json_is_boolean(value)) {
                l2tp_server->data_offset = json_boolean_value(value);
            }
            value = json_object_get(sub, "control-tos");
            if (json_is_number(value)) {
                l2tp_server->control_tos = json_number_value(value);
            }
            value = json_object_get(sub, "data-control-tos");
            if (json_is_number(value)) {
                l2tp_server->data_control_tos = json_number_value(value);
            }
        }
    } else if (json_is_object(section)) {
        fprintf(stderr, "JSON config error: List expected in L2TP server configuration but dictionary found\n");
    }

    /* Traffic Streams Configuration */
    section = json_object_get(root, "streams");
    if (json_is_array(section)) {
        /* Config is provided as array (multiple streams) */
        size = json_array_size(section);
        for (i = 0; i < size; i++) {
            if(!stream_config) {
                ctx->config.stream_config = malloc(sizeof(bbl_stream_config));
                stream_config = ctx->config.stream_config;
            } else {
                stream_config->next = malloc(sizeof(bbl_stream_config));
                stream_config = stream_config->next;
            }
            memset(stream_config, 0x0, sizeof(bbl_stream_config));
            if(!json_parse_stream(ctx, json_array_get(section, i), stream_config)) {
                return false;
            }
        }
    }

    /* Access Line Profiles Configuration */
    section = json_object_get(root, "access-line-profiles");
    if (json_is_array(section)) {
        /* Config is provided as array (multiple access-line-profiles) */
        size = json_array_size(section);
        for (i = 0; i < size; i++) {
            if(!access_line_profile) {
                ctx->config.access_line_profile = malloc(sizeof(bbl_access_line_profile_s));
                access_line_profile = ctx->config.access_line_profile;
            } else {
                access_line_profile->next = malloc(sizeof(bbl_access_line_profile_s));
                access_line_profile = access_line_profile->next;
            }
            memset(access_line_profile, 0x0, sizeof(bbl_access_line_profile_s));
            if(!json_parse_access_line_profile(json_array_get(section, i), access_line_profile)) {
                return false;
            }
        }
    }

    return true;
}

/* bbl_config_load_json
 *
 * This functions is population the BBL context
 * from given JSON configuration file returning
 * true is successfull or false if failed with
 * error message printed to stderr.
 */
bool
bbl_config_load_json (char *filename, bbl_ctx_s *ctx) {
    json_t *root = NULL;
    json_error_t error;
    bool result = false;

    root = json_load_file(filename, 0, &error);
    if(root) {
        result = json_parse_config(root, ctx);
        json_decref(root);
    } else {
        fprintf(stderr, "JSON config error: File %s Line %d: %s\n", filename, error.line, error.text);
    }
    return result;
}

/* bbl_config_load_json
 *
 * This functions is population the BBL context
 * with default configuration values.
 */
void
bbl_config_init_defaults (bbl_ctx_s *ctx) {
    ctx->config.username = (char *)g_default_user;
    ctx->config.password = (char *)g_default_pass;
    ctx->config.tx_interval = 5 * MSEC;
    ctx->config.rx_interval = 5 * MSEC;
    ctx->config.io_slots = 1024;
    ctx->config.io_stream_max_ppi = 32;
    ctx->config.qdisc_bypass = true;
    ctx->config.sessions = 1;
    ctx->config.sessions_max_outstanding = 800;
    ctx->config.sessions_start_rate = 400,
    ctx->config.sessions_stop_rate = 400,
    ctx->config.pppoe_discovery_timeout = 5;
    ctx->config.pppoe_discovery_retry = 10;
    ctx->config.ppp_mru = 1492;
    ctx->config.lcp_conf_request_timeout = 5;
    ctx->config.lcp_conf_request_retry = 10;
    ctx->config.lcp_keepalive_interval = 30;
    ctx->config.lcp_keepalive_retry = 3;
    ctx->config.authentication_timeout = 5;
    ctx->config.authentication_retry = 30;
    ctx->config.ipv4_enable = true;
    ctx->config.arp_timeout = 1;
    ctx->config.arp_interval = 300;
    ctx->config.ipv6_enable = true;
    ctx->config.ipcp_enable = true;
    ctx->config.ipcp_request_ip = true;
    ctx->config.ipcp_request_dns1 = true;
    ctx->config.ipcp_request_dns2 = true;
    ctx->config.ipcp_conf_request_timeout = 5;
    ctx->config.ipcp_conf_request_retry = 10;
    ctx->config.ip6cp_enable = true;
    ctx->config.ip6cp_conf_request_timeout = 5;
    ctx->config.ip6cp_conf_request_retry = 10;
    ctx->config.dhcp_enable = false;
    ctx->config.dhcp_timeout = 5;
    ctx->config.dhcp_retry = 10;
    ctx->config.dhcp_release_interval = 1;
    ctx->config.dhcp_release_retry = 3;
    ctx->config.dhcpv6_enable = true;
    ctx->config.dhcpv6_rapid_commit = true;
    ctx->config.dhcpv6_timeout = 5;
    ctx->config.dhcpv6_retry = 10;
    ctx->config.igmp_autostart = true;
    ctx->config.igmp_version = IGMP_VERSION_3;
    ctx->config.igmp_start_delay = 1;
    ctx->config.igmp_group = 0;
    ctx->config.igmp_group_iter = htobe32(1);
    ctx->config.igmp_source = 0;
    ctx->config.igmp_group_count = 1;
    ctx->config.igmp_zap_wait = true;
    ctx->config.session_traffic_autostart = true;
}
